unit main;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, clMailMessage, clPOP3, clMC, clSMTP, msgfrm,
  clTcpClient, clSyncUtils;

type
  TclMailThread = class(TThread)
  private
    FSynchronizer: TclThreadSynchronizer;
    FStartEvent: THandle;
    FStopEvent: THandle;
    FIsBusy: Boolean;
    FError: string;
    FIsCancelled: Boolean;

    FAsyncMethod: TThreadMethod;
    FPop3: TclPop3;
    FSmtp: TclSmtp;
    FMailMessage: TclMailMessage;
    FMessageNo: Integer;

    FOnReceived: TNotifyEvent;
    FOnSent: TNotifyEvent;
    FOnChecked: TNotifyEvent;
    FOnError: TNotifyEvent;

    procedure AsyncCheckMail;
    procedure AsyncSendMail;
    procedure AsyncReceiveMail;

    procedure DoChecked;
    procedure DoSent;
    procedure DoReceived;
    procedure DoError;
  protected
    procedure Execute; override;
  public
    constructor Create;
    destructor Destroy; override;
    
    procedure CheckMail(APop3: TclPop3; AMessage: TclMailMessage);
    procedure SendMail(ASmtp: TclSmtp; AMessage: TclMailMessage);
    procedure ReceiveMail(APop3: TclPop3; const AMessageNo: Integer; AMessage: TclMailMessage);
    procedure Stop;

    property IsBusy: Boolean read FIsBusy;
    property Error: string read FError;
    property IsCancelled: Boolean read FIsCancelled write FIsCancelled;

    property OnChecked: TNotifyEvent read FOnChecked write FOnChecked;
    property OnSent: TNotifyEvent read FOnSent write FOnSent;
    property OnReceived: TNotifyEvent read FOnReceived write FOnReceived;
    property OnError: TNotifyEvent read FOnError write FOnError;
  end;

  TForm1 = class(TForm)
    Label1: TLabel;
    edtServer: TEdit;
    Label2: TLabel;
    edtUser: TEdit;
    Label3: TLabel;
    edtPassword: TEdit;
    clSMTP1: TclSMTP;
    clPOP31: TclPOP3;
    clMailMessage1: TclMailMessage;
    btnComposeMail: TButton;
    btnCheckMail: TButton;
    lbMessages: TListBox;
    Label4: TLabel;
    btnStop: TButton;
    procedure btnCheckMailClick(Sender: TObject);
    procedure btnComposeMailClick(Sender: TObject);
    procedure lbMessagesDblClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure btnStopClick(Sender: TObject);
    procedure FormCloseQuery(Sender: TObject; var CanClose: Boolean);
  private
    FMailThread: TclMailThread;
    procedure DoOnChecked(Sender: TObject);
    procedure DoOnSent(Sender: TObject);
    procedure DoOnReceived(Sender: TObject);
    procedure DoOnError(Sender: TObject);
  end;

var
  Form1: TForm1;

implementation

{$R *.dfm}

procedure TForm1.btnCheckMailClick(Sender: TObject);
begin
  if FMailThread.IsBusy then Exit;

  clPOP31.Server := edtServer.Text;
  clPOP31.UserName := edtUser.Text;
  clPOP31.Password := edtPassword.Text;

  lbMessages.Items.Clear();
  FMailThread.CheckMail(clPOP31, clMailMessage1);
end;

procedure TForm1.btnComposeMailClick(Sender: TObject);
var
  dlg: TMessageForm;
begin
  if FMailThread.IsBusy then Exit;

  dlg := TMessageForm.Create(nil);
  try
    if dlg.ShowModal() = mrOK then
    begin
      clMailMessage1.BuildMessage(dlg.memBody.Lines.Text, '');
      clMailMessage1.From.FullAddress := dlg.edtFrom.Text;
      clMailMessage1.ToList.EmailAddresses := dlg.edtTo.Text;
      clMailMessage1.Subject := dlg.edtSubject.Text;

      dlg.btnSend.Enabled := True;

      clSMTP1.Server := edtServer.Text;
      clSMTP1.UserName := edtUser.Text;
      clSMTP1.Password := edtPassword.Text;

      FMailThread.SendMail(clSMTP1, clMailMessage1);
    end;
  finally
    dlg.Free();
  end;
end;

procedure TForm1.lbMessagesDblClick(Sender: TObject);
begin
  if FMailThread.IsBusy or (lbMessages.ItemIndex < 0) then Exit;

  clPOP31.Server := edtServer.Text;
  clPOP31.UserName := edtUser.Text;
  clPOP31.Password := edtPassword.Text;

  FMailThread.ReceiveMail(clPOP31, lbMessages.ItemIndex, clMailMessage1);
end;

procedure TForm1.DoOnChecked(Sender: TObject);
begin
  if (clMailMessage1.Subject = '') then
  begin
    lbMessages.Items.Add(clMailMessage1.From.FullAddress);
  end else
  begin
    lbMessages.Items.Add(clMailMessage1.Subject);
  end;
end;

procedure TForm1.DoOnError(Sender: TObject);
begin
  ShowMessage(FMailThread.Error);
end;

procedure TForm1.DoOnReceived(Sender: TObject);
var
  dlg: TMessageForm;
begin
  dlg := TMessageForm.Create(nil);
  try
    dlg.edtFrom.Text := clMailMessage1.From.FullAddress;
    dlg.edtTo.Text := Trim(clMailMessage1.ToList.EmailAddresses);
    dlg.edtSubject.Text := clMailMessage1.Subject;

    dlg.memBody.Lines := clMailMessage1.MessageText;

    dlg.edtFrom.ReadOnly := True;
    dlg.edtTo.ReadOnly := True;
    dlg.edtSubject.ReadOnly := True;
    dlg.memBody.ReadOnly := True;
    dlg.btnSend.Enabled := False;

    dlg.ShowModal();
  finally
    dlg.Free();
  end;
end;

procedure TForm1.DoOnSent(Sender: TObject);
begin
  ShowMessage('The message was sent successfully');
end;

procedure TForm1.btnStopClick(Sender: TObject);
begin
  FMailThread.Stop();
end;

procedure TForm1.FormCreate(Sender: TObject);
begin
  FMailThread := TclMailThread.Create();

  FMailThread.OnChecked := DoOnChecked;
  FMailThread.OnSent := DoOnSent;
  FMailThread.OnReceived := DoOnReceived;
  FMailThread.OnError := DoOnError;
end;

procedure TForm1.FormDestroy(Sender: TObject);
begin
  FMailThread.Free();
end;

procedure TForm1.FormCloseQuery(Sender: TObject; var CanClose: Boolean);
begin
  CanClose := not FMailThread.IsBusy;
  if (not CanClose) then
    ShowMessage('The operation is in progress, press Stop to terminate');
end;

{ TclMailThread }

procedure TclMailThread.AsyncCheckMail;
var
  i: Integer;
begin
  if FPop3.Active then Exit;

  FPop3.Open();
  try
    for i := 1 to FPop3.MessageCount do
    begin
      if IsCancelled then Break;

      FPop3.RetrieveHeader(i, FMailMessage);
      FSynchronizer.Synchronize(DoChecked);
    end;
  finally
    FPop3.Close();
  end;
end;

procedure TclMailThread.AsyncReceiveMail;
begin
  if FPop3.Active then Exit;
  FPop3.Open();
  try
    FPop3.Retrieve(FMessageNo, FMailMessage);
  finally
    FPop3.Close();
  end;
  FSynchronizer.Synchronize(DoReceived);
end;

procedure TclMailThread.AsyncSendMail;
begin
  if FSmtp.Active then Exit;
  FSmtp.Open();
  try
    FSmtp.Send(FMailMessage);
  finally
    FSmtp.Close();
  end;
  FSynchronizer.Synchronize(DoSent);
end;

procedure TclMailThread.CheckMail(APop3: TclPop3; AMessage: TclMailMessage);
begin
  Assert(not FIsBusy);
  IsCancelled := False;

  FAsyncMethod := AsyncCheckMail;
  FPop3 := APop3;
  FSmtp := nil;
  FMailMessage := AMessage;
  FMessageNo := -1;

  FIsBusy := True;
  SetEvent(FStartEvent);
end;

constructor TclMailThread.Create;
begin
  inherited Create(True);
  FSynchronizer := TclThreadSynchronizer.Create();
  FStartEvent := CreateEvent(nil, False, False, nil);
  FStopEvent := CreateEvent(nil, False, False, nil);
  Resume();
end;

destructor TclMailThread.Destroy;
begin
  Stop();
  Terminate();
  SetEvent(FStopEvent);
  WaitForSingleObject(Handle, INFINITE);
  CloseHandle(FStopEvent);
  CloseHandle(FStartEvent);
  FSynchronizer.Free();
  inherited Destroy();
end;

procedure TclMailThread.DoChecked;
begin
  if Assigned(OnChecked) then
  begin
    OnChecked(Self);
  end;
end;

procedure TclMailThread.DoError;
begin
  if Assigned(OnError) then
  begin
    OnError(Self);
  end;
end;

procedure TclMailThread.DoReceived;
begin
  if Assigned(OnReceived) then
  begin
    OnReceived(Self);
  end;
end;

procedure TclMailThread.DoSent;
begin
  if Assigned(OnSent) then
  begin
    OnSent(Self);
  end;
end;

procedure TclMailThread.Execute;
var
  dwResult: DWORD;
  arr: array[0..1] of THandle;
begin
  try
    arr[0] := FStopEvent;
    arr[1] := FStartEvent;
    repeat
      dwResult := WaitForMultipleObjects(2, @arr, FALSE, INFINITE);
      if (dwResult = WAIT_OBJECT_0 + 1) then
      begin
        try
          FAsyncMethod();
        except
          on E: Exception do
          begin
            FError := E.Message;
            FSynchronizer.Synchronize(DoError);
          end;
        end;
        FIsBusy := False;
      end;
    until Terminated or (dwResult = WAIT_OBJECT_0);
  except
    Assert(False);
  end;
end;

procedure TclMailThread.ReceiveMail(APop3: TclPop3; const AMessageNo: Integer; AMessage: TclMailMessage);
begin
  Assert(not FIsBusy);
  IsCancelled := False;

  FAsyncMethod := AsyncReceiveMail;
  FPop3 := APop3;
  FSmtp := nil;
  FMailMessage := AMessage;
  FMessageNo := AMessageNo;

  FIsBusy := True;
  SetEvent(FStartEvent);
end;

procedure TclMailThread.SendMail(ASmtp: TclSmtp; AMessage: TclMailMessage);
begin
  Assert(not FIsBusy);
  IsCancelled := False;

  FAsyncMethod := AsyncSendMail;
  FPop3 := nil;
  FSmtp := ASmtp;
  FMailMessage := AMessage;
  FMessageNo := -1;

  FIsBusy := True;
  SetEvent(FStartEvent);
end;

procedure TclMailThread.Stop;
begin
  IsCancelled := True;
  if (FPop3 <> nil) then
  begin
    FPop3.Close();
  end;
  if (FSmtp <> nil) then
  begin
    FSmtp.Close();
  end;
end;

end.
